/* Definitions of IRC message functions and list of messages.
 *
 * Epona (c) 2000-2001 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 *
 * Based on the original code of Services by Andy Church.
 *
 */

#include "services.h"
#include "messages.h"
#include "language.h"

static char *uplink;

/* List of messages is at the bottom of the file. */

/*************************************************************************/
/*************************************************************************/

static void m_nickcoll(char *source, int ac, char **av)
{
    if (ac < 1)
	return;
    if (!skeleton && !readonly)
	introduce_user(av[0]);
}

/*************************************************************************/

static void m_ping(char *source, int ac, char **av)
{
    if (ac < 1)
	return;
    send_cmd(ServerName, "PONG %s %s", ac>1 ? av[1] : ServerName, av[0]);
}

/*************************************************************************/

static void m_away(char *source, int ac, char **av)
{
    User *u = finduser(source);

    if (u && (ac == 0 || *av[0] == 0))	/* un-away */
	check_memos(u);
}

/*************************************************************************/

static void m_join(char *source, int ac, char **av)
{
    if (ac != 1)
	return;
    do_join(source, ac, av);
}

/*************************************************************************/

static void m_kick(char *source, int ac, char **av)
{
    if (ac != 3)
	return;
    do_kick(source, ac, av);
}

/*************************************************************************/

static void m_kill(char *source, int ac, char **av)
{
    BotInfo *bi;
	
    if (ac != 2)
	return;
    /* Recover if someone kills us. */
    if (stricmp(av[0], s_OperServ) == 0 ||
        stricmp(av[0], s_NickServ) == 0 ||
        stricmp(av[0], s_ChanServ) == 0 ||
        stricmp(av[0], s_MemoServ) == 0 ||
        (s_BotServ && stricmp(av[0], s_BotServ) == 0) ||
        stricmp(av[0], s_HelpServ) == 0 ||
        (s_DevNull && stricmp(av[0], s_DevNull) == 0) ||
        stricmp(av[0], s_GlobalNoticer) == 0
    ) {
	if (!readonly && !skeleton)
	    introduce_user(av[0]);
    } else if (s_BotServ && (bi = findbot(av[0]))) { 
        if (!readonly && !skeleton)
        {
	    introduce_user(av[0]);
	    bot_rejoin_all(bi);
	}
    } else {
	do_kill(source, ac, av);
    }
}

/*************************************************************************/

static void m_mode(char *source, int ac, char **av)
{
    if (*av[0] == '#' || *av[0] == '&') {
	if (ac < 2)
	    return;
	do_cmode(source, ac, av);
    } else {
	if (ac != 2)
	    return;
	do_umode(source, ac, av);
    }
}

/*************************************************************************/

static void m_motd(char *source, int ac, char **av)
{
    FILE *f;
    char buf[BUFSIZE];

    f = fopen(MOTDFilename, "r");
    send_cmd(ServerName, "375 %s :- %s Message of the Day",
		source, ServerName);
   if (f) {
	while (fgets(buf, sizeof(buf), f)) {
	    buf[strlen(buf)-1] = 0;
	    send_cmd(ServerName, "372 %s :- %s", source, buf);
	}
	fclose(f);
    } else {
	send_cmd(ServerName, "372 %s :- MOTD file not found!  Please "
			"contact your IRC administrator.", source);
    }
    send_cmd(ServerName, "376 %s :End of /MOTD command.", source);
}

/*************************************************************************/

static void m_nick(char *source, int ac, char **av)
{
	if (ac != 2) {
#if defined(IRC_BAHAMUT)
# if defined(IRC_SEQUANA)
		User *user = do_nick(source, av[0], av[4], av[5], av[6], av[10], strtoul(av[2], NULL, 10), strtoul(av[7], NULL, 0), strtoul(av[9], NULL, 0), av[8]);
# else
		User *user = do_nick(source, av[0], av[4], av[5], av[6], av[9], strtoul(av[2], NULL, 10), strtoul(av[7], NULL, 0), strtoul(av[8], NULL, 0));
# endif
		if (user)
			set_umode(user, 1, &av[3]);
#elif defined(IRC_UNREAL)
		if (ac == 7) {
			/* For some reasons, Unreal sends this sometimes */
			do_nick(source, av[0], av[3], av[4], av[5], av[6], strtoul(av[2], NULL, 10), 0, "*");
		} else {
			User *user = do_nick(source, av[0], av[3], av[4], av[5], av[9], strtoul(av[2], NULL, 10), strtoul(av[6], NULL, 0), av[8]);
		
			if (user)
				set_umode(user, 1, &av[7]);
		}
#else
# if defined(IRC_ULTIMATE)
		if (ac == 7) {
			do_nick(source, av[0], av[3], av[4], av[5], av[6], strtoul(av[2], NULL, 10), 0);
		} else {
			do_nick(source, av[0], av[3], av[4], av[5], av[7], strtoul(av[2], NULL, 10), strtoul(av[6], NULL, 0));
		}
# else
		do_nick(source, av[0], av[3], av[4], av[5], av[7], strtoul(av[2], NULL, 10), strtoul(av[6], NULL, 0));
# endif
#endif
	} else {
		do_nick(source, av[0], NULL, NULL, NULL, NULL, strtoul(av[1], NULL, 10), 0);
	}	
}

/*************************************************************************/

static void m_part(char *source, int ac, char **av)
{
    if (ac < 1 || ac > 2)
	return;
    do_part(source, ac, av);
}

/*************************************************************************/

static void m_privmsg(char *source, int ac, char **av)
{
    char *s;
    time_t starttime, stoptime;	/* When processing started and finished */
    
    BotInfo *bi;
    ChannelInfo *ci;
    User *u;

    if (ac != 2)
		return;

	u = finduser(source);

	if (!u) {
		log("%s: user record for %s not found", av[1], source);
		notice(av[1], source, getstring(NULL, USER_RECORD_NOT_FOUND));
		return;
    }

    /* Check if we should ignore.  Operators always get through. */
    if (allow_ignore && !is_oper(u)) {
		IgnoreData *ign = get_ignore(source);
		if (ign && ign->time > time(NULL)) {
	    	log("Ignored message from %s: \"%s\"", source, inbuf);
	    	return;
		}
    }

    /* If a server is specified (nick@server format), make sure it matches
     * us, and strip it off. 
     * Don't forget to check that it is not a channel first. */
    if (*av[0] != '#') {
    	s = strchr(av[0], '@');
    	if (s) {
			*s++ = 0;
			if (stricmp(s, ServerName) != 0)
	    		return;
    	}
    }

    starttime = time(NULL);

    if (stricmp(av[0], s_OperServ) == 0) {
		if (is_oper(u)) {
	    	operserv(u, av[1]);
		} else {
			notice_lang(s_OperServ, u, ACCESS_DENIED);
	    
	    	if (WallBadOS)
				wallops(s_OperServ, "Denied access to %s from %s (non-oper)", s_OperServ, u->nick);
		}
    } else if (stricmp(av[0], s_NickServ) == 0) {
		nickserv(u, av[1]);
    } else if (stricmp(av[0], s_ChanServ) == 0) {
		chanserv(u, av[1]);
    } else if (stricmp(av[0], s_MemoServ) == 0) {
		memoserv(u, av[1]);
    } else if (s_BotServ && stricmp(av[0], s_BotServ) == 0) {
		botserv(u, av[1]);
    } else if (stricmp(av[0], s_HelpServ) == 0) {
		notice_help(s_HelpServ, u, HELP_HELP, s_NickServ, s_ChanServ, s_MemoServ);
		if (s_BotServ) 
			notice_help(s_HelpServ, u, HELP_HELP_BOT, s_BotServ);
    } else if (s_BotServ && (bi = findbot(av[0]))) {
    	botmsgs(u, bi, av[1]);
    } else if (s_BotServ && (ci = cs_findchan(av[0]))) {
    	if (!(ci->flags & CI_VERBOTEN) && ci->c && ci->bi) /* Some paranoia checks */
    		botchanmsgs(u, ci, av[1]);	
    }

    /* Add to ignore list if the command took a significant amount of time. */
    if (allow_ignore) {
		stoptime = time(NULL);
		if (stoptime > starttime && *source && !strchr(source, '.'))
	    	add_ignore(source, stoptime-starttime);
    }
}

/*************************************************************************/

static void m_quit(char *source, int ac, char **av)
{
    if (ac != 1)
	return;
    do_quit(source, ac, av);
}

/*************************************************************************/

static void m_server(char *source, int ac, char **av)
{	
	if (!stricmp(av[1], "1"))
		uplink = sstrdup(av[0]);
}

/*************************************************************************/

#if defined(IRC_ULTIMATE) || defined(IRC_UNREAL)

static void m_chghost(char *source, int ac, char **av)
{
    User *u;
    
    if (ac != 2)
		return;
    
    u = finduser(av[0]);
	if (!u) {
	    log("user: CHGHOST for nonexistent user %s", av[0]);
	    return;
	}
	
	change_user_host(u, av[1]);
}

/*************************************************************************/

static void m_sethost(char *source, int ac, char **av)
{
    User *u;
    
    if (ac != 1)
	return;
    
    u = finduser(source);
	if (!u) {
	    log("user: SETHOST for nonexistent user %s", source);
	    return;
	}
	
	change_user_host(u, av[0]);
}

/*************************************************************************/

static void m_chgident(char *source, int ac, char **av)
{
	User *u;
    
    if (ac != 2)
		return;
    
    u = finduser(av[0]);
	if (!u) {
	    log("user: CHGIDENT for nonexistent user %s", av[0]);
	    return;
	}
	
	change_user_username(u, av[1]);
}

/*************************************************************************/

static void m_setident(char *source, int ac, char **av)
{
    User *u;
    
    if (ac != 1)
		return;
    
    u = finduser(source);
	if (!u) {
	    log("user: SETIDENT for nonexistent user %s", source);
	    return;
	}
	
	change_user_username(u, av[0]);
}

/*************************************************************************/

static void m_chgname(char *source, int ac, char **av)
{
	User *u;
    
    if (ac != 2)
		return;
    
    u = finduser(av[0]);
	if (!u) {
	    log("user: CHGNAME for nonexistent user %s", av[0]);
	    return;
	}
	
	change_user_realname(u, av[1]);
}

/*************************************************************************/

static void m_setname(char *source, int ac, char **av)
{
    User *u;
    
    if (ac != 1)
	return;
    
    u = finduser(source);
	if (!u) {
	    log("user: SETNAME for nonexistent user %s", source);
	    return;
	}
	
	change_user_realname(u, av[0]);
}

#endif

/*************************************************************************/

#ifdef IRC_BAHAMUT

static void m_sjoin(char *source, int ac, char **av)
{
	do_sjoin(source, ac, av);
}

#endif

/*************************************************************************/

static void m_stats(char *source, int ac, char **av)
{
	int i;
	NickCore *nc;
	
    if (ac < 1)
	return;
    switch (*av[0]) {
      case 'c':
      case 'C':
      		send_cmd(NULL, "213 %s N *@%s * %s 0 0", source, RemoteServer, uplink);
      		send_cmd(NULL, "213 %s C *@%s * %s %d 0", source, RemoteServer, uplink, RemotePort);
      		send_cmd(NULL, "219 %s %c :End of /STATS report.", source, *av[0]);
    		break;
    	
      case 'l':
	send_cmd(NULL, "211 %s Server SendBuf SentBytes SentMsgs RecvBuf "
		"RecvBytes RecvMsgs ConnTime", source);
	send_cmd(NULL, "211 %s %s %d %d %d %d %d %d %ld", source, RemoteServer,
		read_buffer_len(), total_read, -1,
		write_buffer_len(), total_written, -1,
		start_time);
	send_cmd(NULL, "219 %s l :End of /STATS report.", source);
	break;
	
	  case 'o':
	  case 'O':		
    		for (i = 0; i < RootNumber; i++)
    			send_cmd(NULL, "243 %s O * * %s Root 0", source, ServicesRoots[i]);
			for (i = 0; i < servadmins.count && (nc = servadmins.list[i]); i++) 
				send_cmd(NULL, "243 %s O * * %s Admin 0", source, nc->display);
			for (i = 0; i < servopers.count && (nc = servopers.list[i]); i++) 
				send_cmd(NULL, "243 %s O * * %s Oper 0", source, nc->display);
					
			send_cmd(NULL, "219 %s %c :End of /STATS report.", source, *av[0]);
	break;
	
      case 'u': {
	int uptime = time(NULL) - start_time;
	send_cmd(NULL, "242 %s :Services up %d day%s, %02d:%02d:%02d",
		source, uptime/86400, (uptime/86400 == 1) ? "" : "s",
		(uptime/3600) % 24, (uptime/60) % 60, uptime % 60);
	send_cmd(NULL, "250 %s :Current users: %d (%d ops); maximum %d",
		source, usercnt, opcnt, maxusercnt);
	send_cmd(NULL, "219 %s u :End of /STATS report.", source);
	break;
      } /* case 'u' */

      case 'h':
      case 'i':
      case 'k':
      case 'm':
      case 'y':
	send_cmd(NULL, "219 %s %c :/STATS %c not applicable or not supported.",
		source, *av[0], *av[0]);
	break;
    }
}

/*************************************************************************/

static void m_time(char *source, int ac, char **av)
{
    time_t t;
    struct tm *tm;
    char buf[64];

    time(&t);
    tm = localtime(&t);
    strftime(buf, sizeof(buf), "%a %b %d %H:%M:%S %Y %Z", tm);
    send_cmd(NULL, "391 :%s", buf);
}

/*************************************************************************/

static void m_topic(char *source, int ac, char **av)
{
    if (ac != 4)
	return;
    do_topic(source, ac, av);
}

/*************************************************************************/

void m_version(char *source, int ac, char **av)
{
    if (source)
	send_cmd(ServerName, "351 %s Epona-%s %s :%s -- %s",
			source, version_number, ServerName, version_flags, version_build);
}

/*************************************************************************/

void m_whois(char *source, int ac, char **av)
{
    BotInfo *bi;
    const char *clientdesc;

    if (source && ac >= 1) {
	if (stricmp(av[0], s_NickServ) == 0)
	    clientdesc = desc_NickServ;
	else if (stricmp(av[0], s_ChanServ) == 0)
	    clientdesc = desc_ChanServ;
	else if (stricmp(av[0], s_MemoServ) == 0)
	    clientdesc = desc_MemoServ;
	else if (s_BotServ && stricmp(av[0], s_BotServ) == 0)
	    clientdesc = desc_BotServ;
	else if (stricmp(av[0], s_HelpServ) == 0)
	    clientdesc = desc_HelpServ;
	else if (stricmp(av[0], s_OperServ) == 0)
	    clientdesc = desc_OperServ;
	else if (stricmp(av[0], s_GlobalNoticer) == 0)
	    clientdesc = desc_GlobalNoticer;
	else if (s_DevNull && stricmp(av[0], s_DevNull) == 0)
	    clientdesc = desc_DevNull;
	else if (s_BotServ && (bi = findbot(av[0]))) {
	    /* Bots are handled separately */
	    send_cmd(ServerName, "311 %s %s %s %s * :%s", source, bi->nick, bi->user, bi->host, bi->real);
	    send_cmd(ServerName, "307 %s :%s is a registered nick", source, bi->nick);
	    send_cmd(ServerName, "312 %s %s %s :%s", source, bi->nick, ServerName, ServerDesc);
	    send_cmd(ServerName, "317 %s %s %ld %ld :seconds idle, signon time", source, bi->nick, time(NULL) - bi->lastmsg, start_time);
	    send_cmd(ServerName, "318 %s %s :End of /WHOIS list.", source, bi->nick);
	    return;
	}
	else {
	    send_cmd(ServerName, "401 %s %s :No such service.", source, av[0]);
	    return;
	}
	send_cmd(ServerName, "311 %s %s %s %s * :%s", source, av[0], ServiceUser, ServiceHost, clientdesc);
	send_cmd(ServerName, "312 %s %s %s :%s", source, av[0], ServerName, ServerDesc);
	send_cmd(ServerName, "317 %s %s %ld %ld :seconds idle, signon time", source, av[0], time(NULL) - start_time, start_time);
	send_cmd(ServerName, "318 %s %s :End of /WHOIS list.", source, av[0]);
    }
}

/*************************************************************************/
/*************************************************************************/

Message messages[] = {

    { "401",       NULL },
    { "436",       m_nickcoll },
    { "AWAY",      m_away },
    { "INVITE",    NULL },
    { "JOIN",      m_join },
    { "KICK",      m_kick },
    { "KILL",      m_kill },
    { "MODE",      m_mode },
    { "MOTD",      m_motd },
    { "NICK",      m_nick },
    { "NOTICE",    NULL },
    { "PART",      m_part },
    { "PASS",      NULL },
    { "PING",      m_ping },
    { "PRIVMSG",   m_privmsg },
    { "QUIT",      m_quit },
    { "SERVER",    m_server },
    { "SQUIT",     NULL },
    { "STATS",     m_stats },
    { "TIME",      m_time },
    { "TOPIC",     m_topic },
    { "USER",      NULL },
    { "VERSION",   m_version },
    { "WALLOPS",   NULL },
    { "WHOIS",     m_whois },

    /* DALnet specific messages */
    { "AKILL",     NULL },
    { "GLOBOPS",   NULL },
    { "GNOTICE",   NULL },
    { "GOPER",     NULL },
    { "RAKILL",    NULL },
    { "SILENCE",   NULL },
    { "SVSKILL",   NULL },
    { "SVSMODE",   NULL },
    { "SVSNICK",   NULL },
    { "SVSNOOP",   NULL },
    { "SQLINE",    NULL },
    { "UNSQLINE",  NULL },
    
    /* DreamForge specific messages */
#ifdef IRC_DREAMFORGE
    { "PROTOCTL",  NULL },
#endif
    
    /* Bahamut specific messages */
#ifdef IRC_BAHAMUT
    { "CAPAB", 	   NULL },
    { "SGLINE",    NULL },
    { "SJOIN",     m_sjoin },
    { "SVINFO",    NULL },
    { "SZLINE",    NULL },
    { "UNSGLINE",  NULL },
    { "UNSZLINE",  NULL },
#endif

#ifdef IRC_UNREAL
    { "CHGHOST",   m_chghost },
    { "CHGIDENT",  m_chgident },
    { "CHGNAME",   m_chgname },
	{ "NETINFO",   NULL },
	{ "SETHOST",   m_sethost },
	{ "SETIDENT",  m_setident },
	{ "SETNAME",   m_setname },
#endif

#ifdef IRC_ULTIMATE
    { "CHGHOST",   m_chghost },
    { "CHGIDENT",  m_chgident },
    { "CHGNAME",   m_chgname },
	{ "NETINFO",   NULL },
	{ "SETHOST",   m_sethost },
	{ "SETIDENT",  m_setident },
	{ "SETNAME",   m_setname },
	{ "VCTRL",     NULL },
#endif

    { NULL }

};

/*************************************************************************/

Message *find_message(const char *name)
{
    Message *m;

    for (m = messages; m->name; m++) {
	if (stricmp(name, m->name) == 0)
	    return m;
    }
    return NULL;
}

/*************************************************************************/
